# Lengths (equivalents in meters)
inch = 0.0254
foot = 12 * inch
yard = 3 * foot
mile = 5280 * foot
naut_mile = 1852

# Masses (equivalents in kilograms)
slug = 14.59390
lbm = 0.45359237
short_ton = 2000 * lbm
long_ton = 2240 * lbm
oz = lbm / 16

# Time (equivalents in seconds)
minute = 60
hour = 60 * minute
day = 86400
week = 7 * day
year = 365.25 * day
month = year / 12

# Force (equivalents in Newtons)
lbf = slug * foot

# Speed (equivalents in m/s)
knot = 1.852 * 1000 / hour
mph = mile / hour
fps = foot / 1